package ru.sicamp.sicamphelper.repository

import mu.KLogging
import org.jetbrains.exposed.dao.with
import org.jetbrains.exposed.sql.batchInsert
import org.jetbrains.exposed.sql.transactions.transaction
import org.springframework.beans.factory.annotation.Value
import org.springframework.stereotype.Repository
import ru.sicamp.sicamphelper.db.entity.*
import ru.sicamp.sicamphelper.db.table.Events
import ru.sicamp.sicamphelper.db.table.Groups
import ru.sicamp.sicamphelper.db.table.Rooms
import ru.sicamp.sicamphelper.db.table.Users
import ru.sicamp.sicamphelper.helper.ScheduleHelper
import ru.sicamp.sicamphelper.model.dto.EventDto
import ru.sicamp.sicamphelper.util.InTransaction
import ru.sicamp.sicamphelper.util.checkIds
import java.time.LocalDate

@Repository
class EventRepository(
    @Value("\${exposed.batch-size}")
    private val batchSize: Int,
    private val scheduleHelper: ScheduleHelper,
) {
    suspend fun insertEvent(new: EventDto) = transaction {
        val event = insertSingleEvent(new)
        scheduleHelper.updateSchedulesByEvent(event)
    }

    suspend fun insertEventsBatch(batch: List<EventDto>) = transaction {
        batch.chunked(batchSize).flatMap { insertEvents(it) }
    }

    @InTransaction
    private fun insertEvents(batch: List<EventDto>): List<Schedule> {
        val userIds = batch.map { it.responsible }
        checkIds(Users, userIds)
        val groupIds = batch.mapNotNull { it.group }
        checkIds(Groups, groupIds)
        val roomIds = batch.mapNotNull { it.defaultRoom }
        checkIds(Rooms, roomIds)
        val events = Events.batchInsert(batch) { eventDto ->
            this[Events.name] = eventDto.name
            this[Events.registrationType] = eventDto.registrationType
            this[Events.group] = eventDto.group
            this[Events.eventDate] = eventDto.eventDate
            this[Events.eventTime] = eventDto.eventTime
            this[Events.repeat] = eventDto.repeat
            this[Events.assigneeCount] = eventDto.assigneeCount
            this[Events.status] = eventDto.status
            this[Events.description] = eventDto.description
            this[Events.responsible] = eventDto.responsible
            this[Events.defaultRoom] = eventDto.defaultRoom
        }
            .map { Event.wrapRow(it) }
            .with(
                Event::responsible,
                Event::group,
                Event::defaultRoom
            )

        return events.flatMap {
            scheduleHelper.updateSchedulesByEvent(it)
        }
    }

    @InTransaction
    private fun insertSingleEvent(new: EventDto) = Event.new {
        name = new.name
        registrationType = new.registrationType
        group = new.group?.let {
            Group.findById(it) ?: error("Not found group with id=$it")
        }
        eventDate = new.eventDate
        eventTime = new.eventTime
        repeat = new.repeat
        assigneeCount = new.assigneeCount
        status = new.status
        description = new.description
        responsible = User.findById(new.responsible) ?: error("Not found user with id=${new.responsible}")
        defaultRoom = new.defaultRoom?.let {
            Room.findById(it) ?: error("Not found room with id=${new.defaultRoom}")
        }
    }

    companion object : KLogging()
}