package ru.sicamp.sicamphelper.repository

import org.jetbrains.exposed.dao.with
import org.jetbrains.exposed.sql.*
import org.jetbrains.exposed.sql.javatime.date
import org.jetbrains.exposed.sql.transactions.transaction
import org.springframework.stereotype.Repository
import ru.sicamp.sicamphelper.db.entity.*
import ru.sicamp.sicamphelper.db.table.*
import ru.sicamp.sicamphelper.model.dto.ScheduleDto
import ru.sicamp.sicamphelper.util.InTransaction
import java.time.LocalDate

@Repository
class ScheduleRepository {
    suspend fun updateScheduleById(new: ScheduleDto) = transaction {
        if (new.id == null) {
            error("Id is required for update, but was no no id was provided in scheduleDto: $new")
        }
        val schedule = Schedule.findById(new.id)     ?: error("Not found schedule by id=${new.id}")

        schedule.apply {
            new.dateTime?.let { dateTime = it }
            new.roomId?.let { room = Room.findById(it) }
            new.scheduleDescription?.let { description = it }
            new.responsibleId?.let { responsible = User.findById(it) ?: error("Not found user with id=$it") }
        }
    }

    suspend fun scheduleByGroupId(groupId: Long, date: LocalDate) = transaction {
        val groupSelector = Op.build { GroupRegistrations.group eq groupId }

        return@transaction schedulesByGroupSelector(groupSelector, date)
    }

    suspend fun scheduleByGroupName(name: String, date: LocalDate) = transaction {
        val groupId = Group.find {
            Groups.name eq name
        }.firstOrNull()?.id ?: return@transaction null

        val groupSelector = Op.build { GroupRegistrations.id eq groupId }

        return@transaction schedulesByGroupSelector(groupSelector, date)
    }

    suspend fun scheduleByGateLogin(gateLogin: String, date: LocalDate) = transaction {
        val user = User.find {
            Users.login eq gateLogin
        }.firstOrNull() ?: return@transaction null
        return@transaction internalScheduleByUser(user, date)
    }

    suspend fun scheduleByTgUsername(tgUsername: String, date: LocalDate) = transaction {
        val user = User.find {
            Users.tgUsername eq tgUsername
        }.firstOrNull() ?: return@transaction null
        return@transaction internalScheduleByUser(user, date)
    }

    suspend fun scheduleByUserId(userId: Long, date: LocalDate) = transaction {
        val user = User.findById(userId) ?: return@transaction null
        return@transaction internalScheduleByUser(user, date)
    }

    suspend fun scheduleByUser(user: User, date: LocalDate) = transaction {
        return@transaction internalScheduleByUser(user, date)
    }

    @InTransaction
    private fun internalScheduleByUser(user: User, date: LocalDate): List<Schedule> {
        val userSelector = Op.build { UserRegistrations.id eq user.id }
        val userSchedules = schedulesByUserSelector(userSelector, date)

        val groupSelector = Op.build { GroupRegistrations.group eq user.group.id }
        val groupSchedules = schedulesByGroupSelector(groupSelector, date)

        return userSchedules.plus(groupSchedules)
    }

    @InTransaction
    private fun schedulesByUserSelector(
        userSelector: Op<Boolean>,
        date: LocalDate
    ) = UserRegistrations
        .join(
            Schedules,
            joinType = JoinType.INNER,
            additionalConstraint = {
                UserRegistrations.schedule eq Schedules.id
            }
        )
        .joinEvents()
        .joinRooms()
        .slice(selectedColumns)
        .select {
            userSelector and
                    (Schedules.start.date().date() eq date)
        }
        .map {
            Schedule.wrapRow(it)
        }
        .with(
            Schedule::room,
            Schedule::event,
            Schedule::responsible
        )

        /*.map {
            Room.wrapRow(it)
            Event.wrapRow(it)
            Schedule.wrapRow(it)
        }*/

    private fun schedulesByGroupSelector(
        groupSelector: Op<Boolean>,
        date: LocalDate
    ) = GroupRegistrations
        .join(
            Schedules,
            joinType = JoinType.INNER,
            additionalConstraint = {
                GroupRegistrations.schedule eq Schedules.id
            }
        )
        .joinEvents()
        .joinRooms()
        .select {
            groupSelector and (Schedules.start.date().date() eq date)
        }
        .map {
            Event.wrapRow(it)
            Room.wrapRow(it)
            Schedule.wrapRow(it)
        }

    private fun ColumnSet.joinEvents() = join(
        Events,
        joinType = JoinType.INNER,
        additionalConstraint = {
            Schedules.event eq Events.id
        }
    )

    private fun ColumnSet.joinRooms() = join(
        Rooms,
        joinType = JoinType.INNER,
        additionalConstraint = {
            Schedules.room eq Rooms.id
        }
    )

    private val selectedColumns = listOf(
        Schedules.id,
        Schedules.start,
        Schedules.room,
        Events.id,
        Events.description,
        Events.responsible,
        Rooms.id,
        Rooms.type,
        Rooms.name,
        Rooms.description
    )
}